//go:build !linux
// +build !linux

package configs

// Namespace defines configuration for each namespace.  It specifies an
// alternate path that is able to be joined via setns.
type Namespace struct{}
